{{[banner_logo, "This file is generated by cppbind on 06/23/2021-14:46.\nPlease do not change it manually."]|make_doxygen_comment}}

{% set file_rel_name = path.relpath(file_name, vars.helpers_dir) -%}
package {{vars.helpers_package_prefix}}

typealias StringArray = Array<String>
typealias ObjectArray = Array<Any?>

/**
 * An internal dataclass to keep an information about an object.
 * It is intended to be used by the generated code.
 */
data class CppBindObject(var id : Long, val owner : Boolean = false)

