# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.


from clang.cindex import CursorKind as ElementKind
from clang.cindex import TypeKind, Type
