# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.
from .cppbind_api_parser import (
    TEMPLATE_TYPE_PARAMETER_KEY,
    TEMPLATE_NON_TYPE_PARAMETER_KEY,
    TEMPLATE_NAME_KEY,
    TEMPLATE_INSTANCE_KEY,
    TEMPLATE_KEY
)
